//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#include "ParticleSource.hpp"
#include "ProtectedPtr.hpp"
//------------------------------------------------------------------------------
#include "MyException.hpp"
//------------------------------------------------------------------------------
#include <SPK.h>
#include <SPK_SFML.h>

#include <lua.hpp>
#include <luabind/luabind.hpp>

#include "Global.hpp"
#include "Screen.hpp"
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
using namespace std;
using namespace SPK;
using namespace SPK::SFML;

//------------------------------------------------------------------------------

/** Constructor */
ParticleModel::ParticleModel() : m_lifeTimeMin(0.5), m_lifeTimeMax(1.f)
{

}
//------------------------------------------------------------------------------

/** Constructor */
ParticleModel::ParticleModel(float lifeTimeMin, float lifeTimeMax) : m_lifeTimeMin(0.5), m_lifeTimeMax(1.f)
{
    SetLifeTime(lifeTimeMin, lifeTimeMax);
}

//------------------------------------------------------------------------------
/** Set the particle life time. A random value will be chosen between min and max for each particle. */
void ParticleModel::SetLifeTime(float min, float max)
{
    m_lifeTimeMin = min;
    m_lifeTimeMax = max;
}
//------------------------------------------------------------------------------
/** Add a fixed value for parameter "name".\\n
  * Note: Name can be either "red", "green", "blue", "angle", "size", "mass" or "alpha".
  */
void ParticleModel::AddFixedParam(const std::string &name, float value)
{
	std::pair<int, float[4]> &param = m_paramList[GetParamId(name)];
	param.first = paramFixed;
	param.second[0] = value;
}
//------------------------------------------------------------------------------

/** Add a mutable value for parameter "name". This value will change from start to end during the particle life time.\\n
  * Note: Name can be either "red", "green", "blue", "angle", "size", "mass" or "alpha".
  */
void ParticleModel::AddMutableParam(const std::string &name, float begin, float end)
{
	std::pair<int, float[4]> &param = m_paramList[GetParamId(name)];
	param.first = paramMutable;
	param.second[0] = begin;
	param.second[1] = end;
}
//------------------------------------------------------------------------------

/** Add a random value for parameter "name". For each particle a value will be chosen between min and max.\\n
  * Note: Name can be either "red", "green", "blue", "angle", "size", "mass" or "alpha".
  */
void ParticleModel::AddRandomParam(const std::string &name, float min, float max)
{
	std::pair<int, float[4]> &param = m_paramList[GetParamId(name)];
	param.first = paramRandom;
	param.second[0] = min;
	param.second[1] = max;
}
//------------------------------------------------------------------------------

/** Add a random and mutable value for parameter "name". For each particle a value will be chosen between begin_min and begin_max. At end of life, the value will be between end_min and end_max.\\n
  * Note: Name can be either "red", "green", "blue", "angle", "size", "mass" or "alpha".
  */
void ParticleModel::AddMutableRandomParam(const std::string &name, float begin_min, float begin_max, float end_min, float end_max)
{
	std::pair<int, float[4]> &param = m_paramList[GetParamId(name)];
	param.first = paramMutableRandom;
	param.second[0] = begin_min;
	param.second[1] = begin_max;
	param.second[2] = end_min;
	param.second[3] = end_max;
}
//------------------------------------------------------------------------------
ModelParam ParticleModel::GetParamId(const std::string &name)
{
    if(name == "red")
    {
        return PARAM_RED;
    }
    else if(name == "green")
    {
        return PARAM_GREEN;
    }
    else if(name == "blue")
    {
        return PARAM_BLUE;
    }
    else if(name == "alpha")
    {
        return PARAM_ALPHA;
    }
    else if(name == "angle")
    {
        return PARAM_ANGLE;
    }
    else if(name == "mass")
    {
        return PARAM_MASS;
    }
    else if(name == "size")
    {
        return PARAM_SIZE;
    }
    else
    {
        MY_THROW("Error: Particule param name '" << name << "' unknown.");
    }
}
//------------------------------------------------------------------------------
SPK::Model *ParticleModel::CreateSparkModel() const
{
    // Create flags for spark
    int flagAll = FLAG_NONE;
    int flagMutable = FLAG_NONE;
    int flagRandom = FLAG_NONE;

    for(stockMap::const_iterator it = m_paramList.begin(); it != m_paramList.end(); it++)
    {
        ModelParamFlag flag = static_cast<ModelParamFlag>(1 << it->first);

        flagAll |= flag;

        if(it->second.first == paramMutable || it->second.first == paramMutableRandom)
        {
            flagMutable |= flag;
        }

        if(it->second.first == paramRandom || it->second.first == paramMutableRandom)
        {
            flagRandom |= flag;
        }
    }

    // Crate the model
    Model* model = Model::create(flagAll, flagMutable, flagRandom);

    // Set params
    for(stockMap::const_iterator it = m_paramList.begin(); it != m_paramList.end(); it++)
    {
        const float *valueTbl = it->second.second;

        switch(it->second.first)
        {
            case paramFixed:
                model->setParam(it->first, valueTbl[0]);
                break;
            case paramMutable:
            case paramRandom:
                model->setParam(it->first, valueTbl[0], valueTbl[1]);
                break;
            case paramMutableRandom:
                model->setParam(it->first, valueTbl[0], valueTbl[1], valueTbl[2], valueTbl[3]);
                break;
        }
    }

    // Set life time
    model->setLifeTime(m_lifeTimeMin, m_lifeTimeMax);

    // Ok
    return model;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
/** Constructor */
ParticleEmitter::ParticleEmitter() : m_type(TypeDisk), m_pt1(Vector2D(0, 0)), m_pt2(Vector2D(1, 0)), m_radius(1), m_angleA(0), m_angleB(0), m_direction(1, 0)
{

}
//------------------------------------------------------------------------------
/** Set throw angles */
void ParticleEmitter::SetAngles(float angleA, float angleB)
{
	m_angleA = angleA;
	m_angleB = angleB;
}
//------------------------------------------------------------------------------
/** Set direction */
void ParticleEmitter::SetDirection(const Vector2D &direction)
{
	m_direction = direction;
}
//------------------------------------------------------------------------------

/** Make the emitter as a line */
void ParticleEmitter::MakeLine(const Vector2D &pt1, const Vector2D &pt2)
{
	m_type = TypeLine;
	m_pt1 = pt1;
	m_pt2 = pt2;
}
//------------------------------------------------------------------------------

/** Make the emitter as a circle */
void ParticleEmitter::MakeCircle(const Vector2D &center, float radius)
{
	m_type = TypeCircle;
	m_pt1 = center;
	m_radius = radius;
}
//------------------------------------------------------------------------------

/** Make the emitter as a point */
void ParticleEmitter::MakePoint(const Vector2D &position)
{
	m_type = TypePoint;
	m_pt1 = position;
}
//------------------------------------------------------------------------------

/** Make the emitter as a disk */
void ParticleEmitter::MakeDisk(const Vector2D &center, float radius)
{
	m_type = TypeDisk;
	m_pt1 = center;
	m_radius = radius;
}
//------------------------------------------------------------------------------
SPK::Emitter *ParticleEmitter::CreateEmitter() const
{
    SPK::Emitter *emitter = SphericEmitter::create(SPK::Vector3D(m_direction.x, m_direction.y, 0), m_angleA, m_angleB);

    switch(m_type)
    {
        case TypeLine:
            emitter->setZone(SPK::Line::create(SPK::Vector3D(m_pt1.x, m_pt1.y, 0), SPK::Vector3D(m_pt2.x, m_pt2.y, 0)));
            break;
        case TypeDisk:
            emitter->setZone(SPK::Sphere::create(SPK::Vector3D(m_pt1.x, m_pt1.y, 0), m_radius));
            break;
        case TypeCircle:
            emitter->setZone(SPK::Sphere::create(SPK::Vector3D(m_pt1.x, m_pt1.y, 0), m_radius), false);
            break;
        case TypePoint:
            emitter->setZone(SPK::Point::create(SPK::Vector3D(m_pt1.x, m_pt1.y, 0)));
            break;
        default:
            MY_THROW("Error : Emitter type unknown / not set !");
            break;

    }

    return emitter;
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
ParticleSource::ParticleSource(Group *parent, const std::string &name, const ParticleModel &model, int maxParticleNumber) : Drawable(parent, name), m_sparkGroup(NULL), m_sparkSystem(NULL), m_isAutoDestroy(false)
{
    SFMLPointRenderer *sparkRenderer = SFMLPointRenderer::create();
    sparkRenderer->setBlendMode(sf::Blend::Alpha);

    // Creates the Group
    m_sparkGroup = SPK::Group::create(model.CreateSparkModel(), abs(maxParticleNumber));
    m_sparkGroup->setRenderer(sparkRenderer);
    m_sparkGroup->addEmitter(ParticleEmitter().CreateEmitter()); // Default emitter

    // Creates the System
    m_sparkSystem = SFMLSystem::create(false);
    m_sparkSystem->addGroup(m_sparkGroup);
}
//------------------------------------------------------------------------------
ParticleSource::ParticleSource(const ParticleSource &source) : Drawable(source), m_isAutoDestroy(source.m_isAutoDestroy), m_luaFunction(source.m_luaFunction)
{
    m_sparkSystem = SPK_Copy(SFMLSystem, source.m_sparkSystem);
    m_sparkGroup = m_sparkSystem->getGroup(0);
}
//------------------------------------------------------------------------------
/** Set the event function (\\a fnName should be an existing lua function) that is called when the tank is empty.\\n
  * Function prototype should be : function fnName(thisParticleSource : ParticleSource)
  */
void ParticleSource::SetOnTankEmptyEventFunction(const luabind::object &luaFunction)
{
    if(luaFunction.is_valid() && luabind::type(luaFunction) == LUA_TFUNCTION)
    {
        m_luaFunction = luaFunction;
    }
    else
    {
        m_luaFunction = luabind::object();
    }
}
//------------------------------------------------------------------------------
/** Get the event function that is called when the tank is empty. */
luabind::object ParticleSource::GetOnTankEmptyEventFunction() const
{
    return m_luaFunction;
}
//------------------------------------------------------------------------------
/** Make particle rendered as lines */
void ParticleSource::MakeRenderLines(float length, float width)
{
    SPK::Renderer *backRenderer = m_sparkGroup->getRenderer();

    SFMLLineRenderer* renderer = SFMLLineRenderer::create(length, width);
    //renderer->setBlendMode(sf::Blend::Add);
    renderer->setBlendMode(sf::Blend::Alpha);
    renderer->setGroundCulling(false);

    m_sparkGroup->setRenderer(renderer);

    SPK_Destroy(backRenderer); // We need to destroy it after removing it from the group

    // No more ressources
    ClearResources();
}
//------------------------------------------------------------------------------
/** Make particle rendered as points */
void ParticleSource::MakeRenderPoints(float pointSize)
{
    SPK::Renderer *backRenderer = m_sparkGroup->getRenderer();

    SFMLPointRenderer* renderer = SFMLPointRenderer::create(pointSize);
    renderer->setBlendMode(sf::Blend::Alpha);
    renderer->setGroundCulling(false);

    m_sparkGroup->setRenderer(renderer);

    SPK_Destroy(backRenderer); // We need to destroy it after removing it from the group

    // No more ressources
    ClearResources();
}
//------------------------------------------------------------------------------
/** Make particle rendered as images */
void ParticleSource::MakeRenderImages(Texture &img, const Vector2D &size, bool blendAdd)
{
    SPK::Renderer *backRenderer = m_sparkGroup->getRenderer();

    SFMLQuadRenderer* renderer = SFMLQuadRenderer::create();
    if(blendAdd)
    {
        renderer->setBlendMode(sf::Blend::Add);
    }
    else
    {
        renderer->setBlendMode(sf::Blend::Alpha);
    }
    renderer->setImage(&img);
    renderer->setScale(size.x, size.y);
    renderer->setGroundCulling(false);


    m_sparkGroup->setRenderer(renderer);

    SPK_Destroy(backRenderer); // We need to destroy it after removing it from the group

    // Ressoucres
    AddResource(&img);
}
//------------------------------------------------------------------------------
void ParticleSource::SetEmitter(const ParticleEmitter &emitter)
{
    // Destroy previous
    SPK::Emitter *backEmitter = m_sparkGroup->getEmitter(0);
    m_sparkGroup->removeEmitter(m_sparkGroup->getEmitter(0));
    SPK_Destroy(backEmitter); // We need to destroy it after removing it from the group

    // Add new
    SPK::Emitter *newEmitter = emitter.CreateEmitter();
    m_sparkGroup->addEmitter(newEmitter);

    // Position (needed because there is no ProceedMvt before the next render
    Vector2D pos = GetGlobalPosition();
    // View rect coords
    newEmitter->getZone()->setPosition(SPK::Vector3D(pos.x, pos.y, 0));
}
//------------------------------------------------------------------------------
ParticleSource::~ParticleSource()
{
    if(m_sparkSystem)
    {
        SPK_Destroy(m_sparkSystem);
    }
}
//------------------------------------------------------------------------------
void ParticleSource::Render(sf::RenderTarget &App)
{
    // Check for invalidated pointers
    if(!AreResourcesOk())
    {
        MY_THROW("ParticleSource::Render(): the image used by the ParticleSource was deleted ! Object name: '" << GetName() << "'.")
    }

    // Move the entire system in case of double render
    SPK::Zone* zone = m_sparkGroup->getEmitter(0)->getZone();
    Vector2D zonePos(zone->getPosition().x, zone->getPosition().y);

    m_sparkSystem->SetPosition(GetGlobalPosition() - zonePos);


    if(IsVisible())
    {
        m_sparkSystem->setRenderTarget(&App);
        m_sparkSystem->render();
    }
}
//------------------------------------------------------------------------------
void ParticleSource::ProceedMvt(float time)
{
    Drawable::ProceedMvt(time);

    SPK::Zone* zone = m_sparkGroup->getEmitter(0)->getZone();
    Vector2D pos = GetGlobalPosition();


    // View rect coords
    zone->setPosition(SPK::Vector3D(pos.x, pos.y, 0));

    if(!m_sparkSystem->update(time))
    {
        ProtectedPtr<ParticleSource> ptr = this;

        if(m_luaFunction.is_valid())
        {
            std::vector<luabind::object> argList;
            if(!m_isAutoDestroy) // S'il ne va pas être détruit...
            {
                argList.push_back(luabind::object(Global::GetLuaState(), ProtectedPtr<ParticleSource>(this)));
            }

            GetParentScreenRaw()->RegisterOneCallbackEvent(m_luaFunction, argList);
        }

        // Destroy if needed
        if(m_isAutoDestroy)
        {
            Destroy();
        }
    }
}
//------------------------------------------------------------------------------
/** Set throw angle of the emitter */
void ParticleSource::SetEmitterAngle(float angle)
{
    static_cast<SPK::SphericEmitter*>(m_sparkGroup->getEmitter(0))->setAngles(angle, 0);
}
//------------------------------------------------------------------------------
/** Get the throw angle of the emitter (0 : only in one direction, 2*pi : ain every direction) */
float ParticleSource::GetEmitterAngle() const
{
    return static_cast<const SPK::SphericEmitter*>(m_sparkGroup->getEmitter(0))->getAngleMin();
}
//------------------------------------------------------------------------------
/** Set direction of the emitter (doesn't need to be normalised) */
void ParticleSource::SetEmitterDirection(const Vector2D &direction)
{
    static_cast<SPK::SphericEmitter*>(m_sparkGroup->getEmitter(0))->setDirection(SPK::Vector3D(direction.x, direction.y, 0));
}
//------------------------------------------------------------------------------
/** Set direction of the emitter (doesn't need to be normalised) */
Vector2D ParticleSource::GetEmitterDirection() const
{
    SPK::Vector3D vec3 = static_cast<const SPK::SphericEmitter*>(m_sparkGroup->getEmitter(0))->getDirection();
    return Vector2D(vec3.x, vec3.y);
}
//------------------------------------------------------------------------------
/** Set particle flow (number of particle emitted per second).\\n
  * Value -1 means unlimited.\\n
  * Note: You can't set Flow and tank to -1 (infinite).
  */
void ParticleSource::SetFlow(int flow)
{
    if(flow < -1)
    {
        MY_THROW("Error: flow out of range !");
    }

    m_sparkGroup->getEmitter(0)->setFlow(flow);
}
//------------------------------------------------------------------------------
/** Get particle flow */
int ParticleSource::GetFlow() const
{
    return m_sparkGroup->getEmitter(0)->getFlow();
}
//------------------------------------------------------------------------------

/** Set particle tank (the number of particle that the source can emit).\\n
  * Value -1 means unlimited. When tank is empty, you can add new particle by calling SetTank()\\n
  * Note: You can't set Flow and tank to -1 (infinite).
  */
void ParticleSource::SetTank(int tank)
{
    if(tank < -1)
    {
        MY_THROW("Error: tank out of range !");
    }

    m_sparkGroup->getEmitter(0)->setTank(tank);
}
//------------------------------------------------------------------------------

/** Get particle tank */
int ParticleSource::GetTank() const
{
    return m_sparkGroup->getEmitter(0)->getTank();
}
//------------------------------------------------------------------------------

/** Set particle gravity */
void ParticleSource::SetGravity(const Vector2D &force)
{
    m_sparkGroup->setGravity(SPK::Vector3D(force.x, force.y, 0.f));
}
//------------------------------------------------------------------------------

/** Get gravity */
Vector2D ParticleSource::GetGravity() const
{
    SPK::Vector3D g = m_sparkGroup->getGravity();
    return Vector2D(g.x, g.y);
}
//------------------------------------------------------------------------------

/** Set friction */
void ParticleSource::SetFriction(float friction)
{
    m_sparkGroup->setFriction(friction);
}
//------------------------------------------------------------------------------

/** Get friction */
float ParticleSource::GetFriction() const
{
    return m_sparkGroup->getFriction();
}
//------------------------------------------------------------------------------

/** Set emission force */
void ParticleSource::SetForce(float min, float max)
{
    m_sparkGroup->getEmitter(0)->setForce(min, max);
}
//------------------------------------------------------------------------------

/** Get min emission force */
float ParticleSource::GetMinForce() const
{
    return m_sparkGroup->getEmitter(0)->getForceMin();
}
//------------------------------------------------------------------------------

/** Get max emission force */
float ParticleSource::GetMaxForce() const
{
    return m_sparkGroup->getEmitter(0)->getForceMax();
}
//------------------------------------------------------------------------------
/** Set auto destroy mode : when tank is 0, the ParticleSource should be automatically destroyed. Default is deactivated. */
void ParticleSource::SetAutoDestroy(bool autoDestroy)
{
    m_isAutoDestroy = autoDestroy;
}
//------------------------------------------------------------------------------
/** Check if auto destroy mode is activated */
bool ParticleSource::IsAutoDestroy() const
{
    return m_isAutoDestroy;
}
//------------------------------------------------------------------------------
/** Conpute particle position as if \\a time seconds were elapsed. This is usefull if you to make the user think that the particle system was already running for a long time.\\n
  * \\param time Time which is supposed to be elapsed. Usualy, it should be \\a maxLifeTime.
  */
void ParticleSource::ComputeParticles(float time)
{
    const int steps = 10;
    for(int i = 0; i < steps; i++)
    {
        ProceedMvt(time / float(steps));
    }
}
//------------------------------------------------------------------------------
} // namespace En
//------------------------------------------------------------------------------
