//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#ifndef SOUND_MANAGER_HPP
#define SOUND_MANAGER_HPP

#include "ProtectedPtr.hpp"
#include "Utils.hpp"

#include <SFML/Audio.hpp>
#include <set>
#include <list>
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
class SoundManager;
class SoundPlayer;
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** A SoundManager is used to render a texture. It can be resised. To create a Shape, you should use Group::CreateShape. */
/// LUA_HEADER_COPY_OFF ///

class SoundManager
{
public:
    SoundManager();

    ~SoundManager();

    /// LUA_EXPORT_ON(SoundManager) ///

    /** Play a the sound. */
    ProtectedPtr<SoundPlayer> PlaySound(Sound &sound);

    /** Stop all the sound currently playing in this SoundManager */
    void StopAllSounds();


    /// LUA_EXPORT_OFF ///

    /** Pause all sounds (when the screen lost focus) */
    void PauseAllSounds();

    /** Resume all sounds (when the screen gain focus) */
    void ResumeAllSounds();

    /** Remove a SoundPlayer from the list */
    void UnregisterSound(SoundPlayer *sndToUnregister);

    /** Clear all sounds */
    void ClearAllSounds();


protected:
    typedef std::set<SoundPlayer*> SndTblType;
    //typedef std::list<SoundPlayer*> SndTblType;

    SndTblType m_sndTbl; // Il faut une liste pour ne pas détruire les sons en cas de resize !

    void PlayProtectedSound(SndTblType::iterator it);
};
//------------------------------------------------------------------------------
class SoundPlayer : public sf::Sound, public ProtectedClassBase
{
public:
    /** Create a sound with the given buffer */
    SoundPlayer(SoundManager *parentSndManager, En::Sound *sound);

    /** Destructor. Will unregister the sound */
    ~SoundPlayer();

    /** Get the buffer used for that sound */
    ProtectedPtr<En::Sound> GetSound();


    /// LUA_EXPORT_ON(SoundPlayer, ProtectedPtr<SoundPlayer>) ///
    /** Stop and destoy this SoundPlayer */
    void Stop();

    /** Set the buffer for that sound and protect it */
    void SetSound(En::Sound *buffer);


	/** Set the sound loop state.
	  * This parameter is disabled by default
	  * \param isLoop : True to play in loop, false to play once
	  */
    //LUA// void SetLoop(bool isLoop);

	/** Set the sound pitch.
	  * The default pitch is 1
	  * \param pitch : New pitch
	  */
    //LUA// void SetPitch(float pitch);

	/** Set the sound volume.
	  * The default volume is 100 (max)
	  * \param Volume : volume (in range [0, 100])
	  */
    //LUA// void SetVolume(float volume);

	/** Set the sound position (take 3 values).
	  * The default position is (0, 0, 0)
	  * \param x, y, z : Position of the sound in the world
	  */
    //LUA// void SetPosition(float x, float y, float z);

	/** Set the sound position (take a 3D vector).
	  * The default position is (0, 0, 0)
	  * \param Position : Position of the sound in the world
	  */
    void SetPosition(const Vector3D& position) {sf::Sound::SetPosition(position);}

	/** Set the minimum distance - closer than this distance,
	  * the listener will hear the sound at its maximum volume.
	  * The default minimum distance is 1.0
	  * \param minDistance : New minimum distance for the sound
	  */
    //LUA// void SetMinDistance(float minDistance);

	/** Set the attenuation factor - the higher the attenuation, the
	  * more the sound will be attenuated with distance from listener.
	  * The default attenuation factor 1.0
	  * \param attenuation : New attenuation factor for the sound
	  */
    //LUA// void SetAttenuation(float attenuation);

	/** Set the current playing position of the sound
	  * \param timeOffset : New playing position, expressed in seconds
	  */
    //LUA// void SetPlayingOffset(float timeOffset);

	/** Tell whether or not the sound is looping
	  * \return True if the sound is looping, false otherwise
	  */
    //LUA// bool GetLoop() const;

	/** Get the pitch
	  * \return Pitch value
	  */
    //LUA// float GetPitch() const;

	/** Get the volume
	  * \return Volume value (in range [1, 100])
	  */
    //LUA// float GetVolume() const;

	/** Get the sound position
	  * \return Position of the sound in the world
	  */
    Vector3D GetPosition() const {return sf::Sound::GetPosition();}

	/** Tell if the sound's position is relative to the listener's
	  * position, or if it's absolute
	  * \return True if the position is relative, false if it's absolute
	  */
    //LUA// bool IsRelativeToListener() const;

	/** Get the minimum distance
	  * \return Minimum distance for the sound
	  */
    //LUA// float GetMinDistance() const;

	/** Get the attenuation factor
	  * \return Attenuation factor of the sound
	  */
    //LUA// float GetAttenuation() const;

	/** Get the current playing position of the sound
	  * \return Current playing position, expressed in seconds
	  */
    //LUA// float GetPlayingOffset() const;

    /// LUA_EXPORT_OFF ///

protected:
    SoundManager *m_parentSndManager;
    virtual void SetBuffer(const sf::SoundBuffer &Buffer);

    En::Sound *m_currentSnd; // Protected via callback of En::Sound


private:
    SoundPlayer(SoundPlayer &src); // No copy
};
//------------------------------------------------------------------------------
} // namespace En

#endif // SOUND_MANAGER_HPP
